@@ -1,25 +1,38 @@
+/*-------------------------------------------------------------------------
+ *
+ * AbstractJdbc2ResultSet.java
+ *     This class defines methods of the jdbc2 specification.  This class 
+ *     extends org.postgresql.jdbc1.AbstractJdbc1ResultSet which provides the 
+ *     jdbc1 methods.  The real Statement class (for jdbc2) is 
+ *     org.postgresql.jdbc2.Jdbc2ResultSet
+ *
+ * Copyright (c) 2003, PostgreSQL Global Development Group
+ *
+ * IDENTIFICATION
+ *	  $Header: /cvsroot/pgsql-server/src/interfaces/jdbc/org/postgresql/jdbc2/Attic/AbstractJdbc2ResultSet.java,v 1.25.2.3 2004/02/03 05:25:37 jurka Exp $
+ *
+ *-------------------------------------------------------------------------
+ */
 package org.postgresql.jdbc2;
 
-
+import java.io.CharArrayReader;
+import java.io.InputStream;
+import java.io.IOException;
 import java.math.BigDecimal;
-import java.io.*;
 import java.sql.*;
-import java.text.ParseException;
-import java.text.SimpleDateFormat;
-import java.util.*;
+import java.util.Enumeration;
+import java.util.Hashtable;
+import java.util.Iterator;
+import java.util.StringTokenizer;
+import java.util.Vector;
 import org.postgresql.Driver;
-import org.postgresql.Field;
+import org.postgresql.core.BaseStatement;
+import org.postgresql.core.Field;
 import org.postgresql.core.Encoding;
-import org.postgresql.largeobject.*;
-import org.postgresql.util.PGbytea;
 import org.postgresql.util.PSQLException;
+import org.postgresql.util.PSQLState;
 
 
-/* $Header: /cvsroot/pgsql-server/src/interfaces/jdbc/org/postgresql/jdbc2/AbstractJdbc2ResultSet.java,v 1.10.2.5 2003/03/25 02:23:10 davec Exp $
- * This class defines methods of the jdbc2 specification.  This class extends
- * org.postgresql.jdbc1.AbstractJdbc1ResultSet which provides the jdbc1
- * methods.  The real Statement class (for jdbc2) is org.postgresql.jdbc2.Jdbc2ResultSet
- */
 public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.AbstractJdbc1ResultSet
 {
 
@@ -38,11 +51,10 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 	protected PreparedStatement deleteStatement = null;
 	private PreparedStatement selectStatement = null;
 
-
-
-	public AbstractJdbc2ResultSet(org.postgresql.PGConnection conn, Statement statement, Field[] fields, Vector tuples, String status, int updateCount, long insertOID, boolean binaryCursor)
+  
+	public AbstractJdbc2ResultSet(BaseStatement statement, Field[] fields, Vector tuples, String status, int updateCount, long insertOID, boolean binaryCursor)
 	{
-		super (conn, statement, fields, tuples, status, updateCount, insertOID, binaryCursor);
+		super (statement, fields, tuples, status, updateCount, insertOID, binaryCursor);
 	}
 
 	public java.net.URL getURL(int columnIndex) throws SQLException
@@ -137,11 +149,18 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 
 			default:
 				String type = field.getPGType();
+
 				// if the backend doesn't know the type then coerce to String
 				if (type.equals("unknown"))
 				{
 					return getString(columnIndex);
 				}
+                                // Specialized support for ref cursors is neater.
+                                else if (type.equals("refcursor"))
+                                {
+                                        String cursorName = getString(columnIndex);
+                                        return statement.createRefCursorResultSet(cursorName);
+                                }
 				else
 				{
 					return connection.getObject(field.getPGType(), getString(columnIndex));
@@ -235,13 +254,15 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 
 	public java.sql.Array getArray(int i) throws SQLException
 	{
+		checkResultSet( i );
+
 		wasNullFlag = (this_row[i - 1] == null);
 		if (wasNullFlag)
 			return null;
 
 		if (i < 1 || i > fields.length)
-			throw new PSQLException("postgresql.res.colrange");
-		return (java.sql.Array) new org.postgresql.jdbc2.Array( connection, i, fields[i - 1], (java.sql.ResultSet) this );
+			throw new PSQLException("postgresql.res.colrange", PSQLState.INVALID_PARAMETER_VALUE);
+		return (java.sql.Array) new org.postgresql.jdbc2.Array( connection, i, fields[i - 1], this );
 	}
 
 
@@ -368,15 +389,6 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 	}
 
 
-	public int getFetchSize() throws SQLException
-	{
-		// In this implementation we return the entire result set, so
-		// here return the number of rows we have. Sub-classes can return a proper
-		// value
-		return rows.size();
-	}
-
-
 	public Object getObject(String columnName, java.util.Map map) throws SQLException
 	{
 		return getObject(findColumn(columnName), map);
@@ -403,7 +415,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 	public Ref getRef(int i) throws SQLException
 	{
 		//The backend doesn't yet have SQL3 REF types
-		throw new PSQLException("postgresql.psqlnotimp");
+		throw new PSQLException("postgresql.psqlnotimp", PSQLState.NOT_IMPLEMENTED);
 	}
 
 
@@ -421,7 +433,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 	// This one needs some thought, as not all ResultSets come from a statement
 	public Statement getStatement() throws SQLException
 	{
-		return statement;
+		return (Statement) statement;
 	}
 
 
@@ -478,9 +490,14 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 
 	public boolean previous() throws SQLException
 	{
-		if (--current_row < 0)
+		if (current_row-1 < 0) {
+			current_row = -1;
 			return false;
+		} else {
+			current_row--;
+		}
 		this_row = (byte[][]) rows.elementAt(current_row);
+		rowBuffer = new byte[this_row.length][];
 		System.arraycopy(this_row, 0, rowBuffer, 0, this_row.length);
 		return true;
 	}
@@ -495,14 +512,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 
 	public void setFetchDirection(int direction) throws SQLException
 	{
-		throw new PSQLException("postgresql.psqlnotimp");
-	}
-
-
-	public void setFetchSize(int rows) throws SQLException
-	{
-		// Sub-classes should implement this as part of their cursor support
-		throw org.postgresql.Driver.notImplemented();
+		throw new PSQLException("postgresql.psqlnotimp", PSQLState.NOT_IMPLEMENTED);
 	}
 
 
@@ -513,7 +523,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		{
 			doingUpdates = false;
 
-			clearRowBuffer();
+			clearRowBuffer(true);
 		}
 	}
 
@@ -656,7 +666,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 			this_row = rowBuffer;
 
 			// need to clear this in case of another insert
-			clearRowBuffer();
+			clearRowBuffer(false);
 
 
 		}
@@ -671,10 +681,15 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 			throw new PSQLException( "postgresql.updateable.notupdateable" );
 		}
 
-		this_row = (byte[][]) rows.elementAt(current_row);
+		if (current_row < 0) {
+			this_row = null;
+			rowBuffer = null;
+		} else {
+			this_row = (byte[][]) rows.elementAt(current_row);
 
-		rowBuffer = new byte[this_row.length][];
-		System.arraycopy(this_row, 0, rowBuffer, 0, this_row.length);
+			rowBuffer = new byte[this_row.length][];
+			System.arraycopy(this_row, 0, rowBuffer, 0, this_row.length);
+		}
 
 		onInsertRow = false;
 		doingUpdates = false;
@@ -696,7 +711,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 
 
 		// make sure the underlying data is null
-		clearRowBuffer();
+		clearRowBuffer(false);
 
 		onInsertRow = true;
 		doingUpdates = false;
@@ -704,12 +719,17 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 	}
 
 
-	private synchronized void clearRowBuffer()
+	private synchronized void clearRowBuffer(boolean copyCurrentRow)
 	throws SQLException
 	{
 		// rowBuffer is the temporary storage for the row
 		rowBuffer = new byte[fields.length][];
 
+		// inserts want an empty array while updates want a copy of the current row
+		if (copyCurrentRow) {
+			System.arraycopy(this_row, 0, rowBuffer, 0, this_row.length);
+		}
+
 		// clear the updateValues hashTable for the next set of updates
 		updateValues.clear();
 
@@ -754,11 +774,10 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		}
 		catch (IOException ie)
 		{
-			throw new PSQLException("postgresql.updateable.ioerror" + ie);
+			throw new PSQLException("postgresql.updateable.ioerror", ie);
 		}
 
 		updateValue(columnIndex, theData);
-
 	}
 
 
@@ -788,10 +807,9 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		}
 		catch (IOException ie)
 		{
-			throw new PSQLException("postgresql.updateable.ioerror" + ie);
+			throw new PSQLException("postgresql.updateable.ioerror", ie);
 		}
 		updateValue(columnIndex, theData);
-
 	}
 
 
@@ -836,7 +854,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		}
 		catch (IOException ie)
 		{
-			throw new PSQLException("postgresql.updateable.ioerror" + ie);
+			throw new PSQLException("postgresql.updateable.ioerror", ie);
 		}
 		updateValue(columnIndex, theData);
 	}
@@ -972,6 +990,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 			}
 
 			rows.setElementAt( rowBuffer, current_row );
+			this_row = rowBuffer;
 			if ( Driver.logDebug )
 				Driver.debug("done updates");
 
@@ -1013,7 +1032,9 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 				{
 
 					String column = (String) columns.nextElement();
-					updateSQL.append( column + "= ?");
+					updateSQL.append("\"");
+					updateSQL.append( column );
+					updateSQL.append("\" = ?");
 
 					if ( i < numColumns - 1 )
 					{
@@ -1030,7 +1051,9 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 				{
 
 					PrimaryKey primaryKey = ((PrimaryKey) primaryKeys.get(i));
-					updateSQL.append(primaryKey.name).append("= ?");
+					updateSQL.append("\"");
+					updateSQL.append(primaryKey.name);
+					updateSQL.append("\" = ?");
 
 					if ( i < numKeys - 1 )
 					{
@@ -1071,7 +1094,6 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 				rows.setElementAt( rowBuffer, current_row );
 				if ( Driver.logDebug )
 					Driver.debug("done updates");
-				// make sure next one doesn't see remnants of this one
 				updateValues.clear();
 				doingUpdates = false;
 			}
@@ -1266,22 +1288,6 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 	}
 
 
-	private int _findColumn( String columnName )
-	{
-		int i;
-
-		final int flen = fields.length;
-		for (i = 0; i < flen; ++i)
-		{
-			if (fields[i].getName().equalsIgnoreCase(columnName))
-			{
-				return (i + 1);
-			}
-		}
-		return -1;
-	}
-
-
 	/**
 	 * Is this ResultSet updateable?
 	 */
@@ -1318,12 +1324,18 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 
 
 		usingOID = false;
-		int oidIndex = _findColumn( "oid" );
+		int oidIndex = 0;
+		try {
+		  oidIndex = findColumn( "oid" );
+		} catch (SQLException l_se) {
+			//Ignore if column oid isn't selected
+		}
 		int i = 0;
 
 
 		// if we find the oid then just use it
 
+		//oidIndex will be >0 if the oid was in the select list
 		if ( oidIndex > 0 )
 		{
 			i++;
@@ -1333,21 +1345,13 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		else
 		{
 			// otherwise go and get the primary keys and create a hashtable of keys
-			// if the user has supplied a quoted table name
-			// remove the quotes, but preserve the case.
-			// otherwise fold to lower case.
-			String quotelessTableName;
-			if (tableName.startsWith("\"") && tableName.endsWith("\"")) {
-				quotelessTableName = tableName.substring(1,tableName.length()-1);
-			} else {
-				quotelessTableName = tableName.toLowerCase();
-			}
-			java.sql.ResultSet rs = ((java.sql.Connection) connection).getMetaData().getPrimaryKeys("", "", quotelessTableName);
-
+			String[] s = quotelessTableName(tableName);
+			String quotelessTableName = s[0];
+			String quotelessSchemaName = s[1];
+			java.sql.ResultSet rs = ((java.sql.Connection) connection).getMetaData().getPrimaryKeys("", quotelessSchemaName, quotelessTableName);
 			for (; rs.next(); i++ )
 			{
 				String columnName = rs.getString(4);	// get the columnName
-
 				int index = findColumn( columnName );
 
 				if ( index > 0 )
@@ -1377,7 +1381,56 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		return updateable;
 	}
 
-
+	/** Cracks out the table name and schema (if it exists) from a fully
+	 * qualified table name.
+	 * @param fullname string that we are trying to crack.	Test cases:<pre>
+	 * Table: table ()
+	 * "Table": Table ()
+	 * Schema.Table: table (schema)
+	 * "Schema"."Table": Table (Schema)
+	 * "Schema"."Dot.Table": Dot.Table (Schema)
+	 * Schema."Dot.Table": Dot.Table (schema)
+	 * </pre>
+	 * @return String array with element zero always being the tablename and
+	 * element 1 the schema name which may be a zero length string.
+	 */
+	public static String[] quotelessTableName(String fullname) {
+		StringBuffer buf = new StringBuffer(fullname);
+		String[] parts = new String[] {null, ""};
+		StringBuffer acc = new StringBuffer();
+		boolean betweenQuotes = false;
+		for (int i = 0; i < buf.length(); i++) {
+			char c = buf.charAt(i);
+			switch (c) {
+				case '"':
+					if ((i < buf.length() - 1) && (buf.charAt(i+1) == '"')) {
+						// two consecutive quotes - keep one
+						i++;
+						acc.append(c);	// keep the quote
+					}
+					else {	// Discard it
+						betweenQuotes = !betweenQuotes;
+					}
+					break;
+				case '.':
+					if (betweenQuotes) {   // Keep it
+						acc.append(c);
+					}
+					else {	   // Have schema name
+						parts[1] = acc.toString();
+						acc = new StringBuffer();
+					}
+					break;
+				default:
+					acc.append((betweenQuotes) ? c : Character.toLowerCase(c));
+					break;
+			}
+		}
+		// Always put table in slot 0
+		parts[0] = acc.toString();
+		return parts;
+	}
+ 
 	public void parseQuery()
 	{
 		String[] l_sqlFragments = ((AbstractJdbc2Statement)statement).getSqlFragments();
@@ -1417,7 +1470,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		while ( columns.hasMoreElements() )
 		{
 			String columnName = (String) columns.nextElement();
-			int columnIndex = _findColumn( columnName ) - 1;
+			int columnIndex = findColumn( columnName ) - 1;
 
 			Object valueObject = updateValues.get(columnName);
 			if (valueObject instanceof NullObject) {
@@ -1459,7 +1512,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 	}
 
 
-	public void setStatement(Statement statement)
+	public void setStatement(BaseStatement statement)
 	{
 		this.statement = statement;
 	}
